const DefaultSize = 512

export const Rgba8 = {
    r: new Uint8Array(DefaultSize),
    g: new Uint8Array(DefaultSize),
    b: new Uint8Array(DefaultSize),
    a: new Uint8Array(DefaultSize),
}

export const Range = {
    min: new Float32Array(DefaultSize),
    max: new Float32Array(DefaultSize)
}

export const UVector2 = {
    x: new Uint32Array(DefaultSize),
    y: new Uint32Array(DefaultSize)
}

export const UVector3 = {
    ...UVector2,
    z: new Uint32Array(DefaultSize)
}

export const Vector2 = {
    x: new Float32Array(DefaultSize),
    y: new Float32Array(DefaultSize),
}

export const Vector3 = {
    ...Vector2,
    z: new Float32Array(DefaultSize)
}

export const Quaternion = {
    ...Vector3,
    w: new Float32Array(DefaultSize)
}

export const Transform = {
    translation: Vector3,
    rotation: Quaternion,
    scale: Vector3
}

export const Viewport = {
    position: UVector2,
    size: UVector2,
    depth: Range
}

export const RenderTarget = []

class RenderGraph {
    #nodes = new Map()
    #edges = new Map()

    add(name, render, dependencies = []) {
        this.node.set(name, render)
        this.edges.set(name, dependencies)
    }

    update(world) {
        const executed = new Set()
        const order = this.sort()

        if(!order) {
            return
        }

        for(const node of order) {
            const render = this.nodes.get(node)
            render(world)
            executed.add(node)
        }
    }

    sort() {
        const visited = new Set()
        const stack = []
        const cycles = new Set()

        const visit = node => {
            if(cycles.has(node)) {
                return false
            }

            if (visited.has(node)) {
                return true
            }

            visited.add(node)
            cycles.add(node)

            for(const dep of this.dependentsFor(name)) {
                if (!visit(dep)) {
                    return false
                }
            }

            cycles.delete(name)
            stack.push(node)
        }

        for(const node in this.nodes) {
            if(!visited.has(node) && !visit(node)) {
                return
            }
        }

        return stack.reverse()
    }

    dependentsFor(name) {
        return this.edges.values()
            .filter(deps => deps.includes(name))
    }
}

export const Camera = {
    viewport: Viewport,
    order: new Int32Array(DefaultSize),
    clearColor: Rgba8,
    renderTarget: RenderTarget
}

export const Renderable = {}

const Topology = Object.freeze({
    Points: 0,
    Lines: 1,
    LineStrip: 2,
    LineLoop: 3,
    Triangles: 4,
    TriangleStrip: 5,
    TriangleFan: 6
})

export const VertexPositions = new Float32Array(1024)
export const Indices = new Uint16Array(1024)

export const Mesh = {
    topology: new Uint8Array(DefaultSize),
    positions: new Uint32Array(DefaultSize),
    indices: new Uint32Array(DefaultSize),
    vao: new Uint32Array(DefaultSize),
    drawMode: new Uint32Array(DefaultSize),
    vertexCount: new Uint32Array(DefaultSize),
    indexCount: new Uint32Array(DefaultSize)
}

export const Sprite = {
    data: [],
    height: new Uint32Array(DefaultSize),
    width: new Uint32Array(DefaultSize)
}

