class GLObject {
    #gl

    constructor(gl) {
        this.#gl = gl
    }
}

export class ShaderError extends Error {
    constructor(message) {
        super(message)
    }
}

export class Shader extends GLObject {
    #gl
    #shader
    #type

    constructor(gl, type, source) {
        super(gl)
        const shader = gl.createShader(type)
        gl.shaderSource(shader, source)
        gl.compileShader(shader)

        if (!gl.getShaderParameter(shader, gl.COMPILE_STATUS)) {
            gl.deleteShader(shader)
            throw new ShaderError(gl.getShaderInfoLog(shader))
        }

        this.#shader = shader
        this.#type = type
    }

    static vertex(gl, source) {
        return new Shader(gl, gl.VERTEX_SHADER, source)
    }

    static fragment(gl, source) {
        return new Shader(gl, gl.FRAGMENT_SHADER, source)
    }

    attach(program) {
        this.#gl.attachShader(program, this.#shader)
    }
}

export class ProgramError extends Error {
    constructor(message) {
        super(message)
    }
}

export class Program {
    #program

    constructor(gl, vertexShader, fragmentShader) {
        super(gl)

        const program = gl.createProgram()

        vertexShader.attach(program)
        fragmentShader.attach(program)

        gl.linkProgram(program)

        if (!gl.getProgramParameter(program, gl.LINK_STATUS)) {
            gl.deleteProgram(program)
            throw new ProgramError(gl.getProgramInfoLog(program))
        }

        this.#program = program

    }

    static fromSource(gl, vertexSource, fragmentSource) {
        return new Program(
            gl,
            Shader.vertex(gl, vertexSource),
            Shader.fragment(gl, fragmentSource)
        )
    }
}

export const DataType = {
    i8: 5120,
    u8: 5121,
    i16: 5122,
    u16: 5123,
    f16: 5131,
    f32: 5126,
    i32: 5124,
    u32: 5125
}

export class Buffer extends GLObject {
    #buffer
    #target

    constructor(gl, target = gl.ARRAY_BUFFER) {
        super(gl)

        const buffer = gl.createBuffer()

        this.#buffer = buffer
        this.#target = target
    }

    get size() {
        return this.#gl.getBufferParameter(
            this.#target,
            this.#gl.BUFFER_SIZE
        )
    }

    get usage() {
        return this.#gl.getBufferParameter(
            this.#target,
            this.#gl.BUFFER_USAGE
        )
    }

    bind() {
        this.#gl.bindBuffer(this.#target, this.#buffer)
    }

    set(value, usage) {
        if(!usage) {
            console.warn('WARN: Buffer.set called without "usage" argument. Defaulting to DYNAMIC_DRAW')
            usage = gl.DYNAMIC_DRAW
        }

        this.#gl.bufferData(
            this.#target,
            value,
            usage
        )
    }
}

export class BufferSchema {
    type
    #size
    normalize
    stride
    offset

    constructor(type, { size = 4, normalize = false, stride = 0, offset = 0 } = {}) {
        this.type = type
        this.#size = size
        this.normalize = normalize
        this.stride = stride
        this.offset = offset
    }

    get size() {
        return this.#size
    }

    set size(value) {
        if (value > 0 && value < 5) {
            this.#size = value
        } else {
            throw new RangeError("BufferSchema size must be between 1 and 4")
        }
    }

    bind(gl, location) {
        gl.vertexAttribPointer(
            location,
            this.size,
            this.type,
            this.normalize,
            this.stride,
            this.offset
        )
    }
}

export class Attribute extends GLObject {
    #name
    #buffer
    #program
    #location
    #schema

    constructor(gl, program, name, buffer, schema) {
        super(gl)

        this.#name = name
        this.#buffer = buffer
        this.#program = program
        this.#schema = schema
        this.#location = gl.getAttribLocation(program, name)
    }

    get buffer() {
        return this.#buffer
    }

    get location() {
        return this.#location
    }

    bind() {
        this.#gl.enableVertexAttribArray(this.#location)
        this.#buffer.bind()
        this.#schema.bind(this.#gl, this.#location)
    }
}

export const UniformType = {
    Int: 'i',
    IntVec: 'iv',
    UInt: 'ui',
    UIntVec: 'uiv',
    Float: 'f',
    FloatVec: 'fv'
}

// TODO: clean this api up
export class Uniform extends GLObject {
    #name
    #program
    #location
    #length
    #type
    #fn

    constructor(gl, program, name, length, type) {
        super(gl)
        this.#name = name
        this.#program = program
        this.#location = gl.getUniformLocation(program, name)
        this.#length = length
        this.#type = type
        this.#setFn(gl, length, type)
    }

    #setFn(gl, length, type) {
        this.#fn = gl[`uniform${length}${type}`].bind(gl)
    }

    get length() {
        return this.#length
    }

    set length(value) {
        this.#length = value
        this.#setFn(this.#gl, value, this.#type)
    }

    get type() {
        return this.#type
    }

    set type(value) {
        this.#type = value
        this.#setFn(this.#gl, this.#length, value)
    }

    // TODO: add data validation
    set(...args) {
        this.#fn(this.#location, ...args)
    }
}

export class Viewport extends GLObject {
    constructor(gl, clearColor = [0, 0, 0, 1]) {
        super(gl)
        this.clearColor = clearColor
    }

    get clearColor() {
        return this.#gl.getParameter(this.#gl.COLOR_CLEAR_VALUE)
    }

    set clearColor(value) {
        this.#gl.clearColor(...value)
    }

    get MaxViewportDimensions() {
        return this.#gl.getParameter(this.#gl.MAX_VIEWPORT_DIMS)
    }

    get bounds() {
        return this.#gl.getParameter(this.#gl.VIEWPORT)
    }

    resize(x, y, width, height) {
        this.#gl.viewport(x, y, width, height)
    }

    clear() {
        this.#gl.clear(this.#gl.COLOR_BUFFER_BIT)
    }
}

